"""Information about urpatimeout module."""
__title__ = "urpautils"
__description__ = "Module containig universal functions used with UltimateRPA robots"
__version__ = "0.1.3"
__author__ = "anerold"
__author_email__ = "support@ultimaterpa.cz"
__url__ = "https://github.com/ultimaterpa/urpautils"
__license__ = "MIT"
